/* Transparent SOCKS proxy for use with Linux kernel.
 *
    Copyright (C) 2003-4 Regents of the University of California; Mike Fisk <mfisk@lanl.gov>
    Copyright (C) 1999  V. Ganesh <ganesh@vxindia.veritas.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <assert.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <pwd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <linux/netfilter_ipv4.h>

#include <socks.h>

/* paranoia, I know. the socks libraries should take care of 
   non-socks sockets, but ...
*/
#define real_listen		listen
#define real_bind		bind
#define real_accept		accept
#define real_connect		connect
#define real_getsockname	getsockname
#define real_getpeername	getpeername
#define real_setsockopt		setsockopt
#define real_getsockopt		getsockopt
#define real_read		read
#define real_write		write
#define real_close		close
#define real_shutdown		shutdown

#define MAXTRIES		5
#define SLEEP_BETWEEN_TRIES	3
#define BUFSIZE			16384
#define CONNECT_TIMEOUT		60
#define WRITE_TIMEOUT		300
#define IDLE_TIMEOUT		0

void socks_serve(int, char **);
void pass_data(int, int);
void alarm_handler(int);

static void logmsg(char *);
static void logerr(char *);

int isdaemon = 1;
short bindport = 1211;

void register_pid() {
	FILE * fh;
	fh = fopen("/var/run/transocks.pid", "w");
	assert(fh);
	fprintf(fh, "%d", getpid());
	fclose(fh);
}

main(argc, argv)
	int	argc;
	char	**argv;
{
	struct sockaddr_in	addr;
	int			addrlen = sizeof(addr);
	int			listenfd, newfd;
	int			newpid;
	char			c;
	int			on = 1;

	while ((c = getopt(argc, argv, "fp:")) != EOF) {
		switch(c) {
		case 'f':
			isdaemon = 0;
			break;
		case 'p':
			if (!atoi(optarg)) {
				fprintf(stderr, "Invalid port %s\n", optarg);
				exit(-1);
			}
			bindport = atoi(optarg);
			break;
		case '?':
			fprintf(stderr, "Usage: %s [-f] [-p port]\n", argv[0]);
			exit(1);
			break;
		}

	}

	if (isdaemon) {
		if ((newpid = fork()) < 0) {
			perror("fork");
			exit(-1);
		}
		if (newpid) { /* parent */
			exit(0);
		}
		fclose(stdin);
		fclose(stdout);
		fclose(stderr);
		setsid();
		setpgrp();
		openlog("transocks", LOG_PID, LOG_DAEMON);
	}

	listenfd = socket(AF_INET, SOCK_STREAM, 0);
	if (listenfd < 0) {
		logerr("socket");
		exit(-1);
	}

	bzero(&addr, sizeof (addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(bindport);

	real_setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	if (real_bind(listenfd, (struct sockaddr *)&addr, sizeof (addr)) < 0) {
		logerr("bind");
		exit(-1);
	}

	if (real_listen(listenfd, SOMAXCONN) < 0) {
		logerr("listen");
		exit(-1);
	}

	signal(SIGCHLD, SIG_IGN);

	/* Ready for business */

	if (isdaemon) register_pid();

	for (;;) {
		if ((newfd = real_accept(listenfd, (struct sockaddr *)&addr, &addrlen)) < 0) {
			logerr("accept");
			exit(-1);
		}

		/* Spawn a new child to handle this client connection */
		newpid = fork();
		if (newpid == 0) {
			close(listenfd);
			socks_serve(newfd, argv);
			exit(0);
		} else if (newpid < 0) {
			logerr("fork");
			real_close(newfd);
			//exit(-1);
		} else {
			real_close(newfd);
		}

	}
}

void
socks_serve(newfd, argv)
	int	newfd;
	char ** argv;
{
	struct sockaddr_in	dstaddr;
	int			dstlen = sizeof(dstaddr);
	int			tries = MAXTRIES;
	int			clientfd, serverfd;
	char			msg[1024];
	struct sockaddr_in 	srcaddr;
	int 			srclen = sizeof(srcaddr);
	char 			srcaddrstr[INET_ADDRSTRLEN];
	char 			dstaddrstr[INET_ADDRSTRLEN];

	if (isdaemon) {
		closelog();
		openlog("transocks", LOG_PID, LOG_DAEMON);
	}

	clientfd = newfd;
	serverfd = socket(AF_INET, SOCK_STREAM, 0);
	if (serverfd < 0) {
		logerr("socket");
		exit(-1);
	}
	
	signal(SIGHUP, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, alarm_handler);
	alarm(CONNECT_TIMEOUT);

#ifdef IPCHAINS
	real_getsockname(clientfd, (struct sockaddr *)&dstaddr, &dstlen); 
#else /* IPTABLES */
	real_getsockopt(clientfd, SOL_IP, SO_ORIGINAL_DST, (struct sockaddr *)&dstaddr, &dstlen);
#endif

	real_getpeername(clientfd, (struct sockaddr *)&srcaddr, &srclen); 
	inet_ntop(AF_INET, &srcaddr.sin_addr, srcaddrstr, INET_ADDRSTRLEN);
	inet_ntop(AF_INET, &dstaddr.sin_addr, dstaddrstr, INET_ADDRSTRLEN);

	snprintf(msg, 1023, "%s:%d -> %s:%d", 
			srcaddrstr, ntohs(srcaddr.sin_port),
			dstaddrstr, ntohs(dstaddr.sin_port));
	logmsg(msg);

	if (isdaemon) {
		snprintf(argv[0], 1023, "transocks %s", msg);
	}

	/* deal with a socks server that intermittently refuses connections */
	while (tries--) {
		if (connect(serverfd, (struct sockaddr *)&dstaddr, sizeof (dstaddr)) == 0) {
			alarm(0);
			pass_data(clientfd, serverfd);
			return;
		}
		sleep(SLEEP_BETWEEN_TRIES);
	}
	logerr("connect");
	return;
}

void
pass_data(clientfd, serverfd)
	int	clientfd;
	int	serverfd;
{
	char	buf[BUFSIZE];
	int	buflen;
	fd_set	readfds;
	int	fdmax;
	int	close_state = 0;
	int	nwritten;
	char	*bp;

#define C_TO_S 1
#define S_TO_C 2

	fdmax = (clientfd > serverfd) ? clientfd : serverfd;
	fdmax++;

	/* a possibly misguided attempt is made to implement
	   shutdown semantics. seems to be harmless though
	*/
	while (1) {
		FD_ZERO(&readfds);

		/* Select for read only on connections that we haven't
		 * already read to the end of. */
		if (! (close_state & C_TO_S)) FD_SET(clientfd, &readfds);
		if (! (close_state & S_TO_C)) FD_SET(serverfd, &readfds);

		if (select(fdmax, &readfds, NULL, NULL, NULL) < 0) {
			logerr("select");
			break;
		}

		if (FD_ISSET(clientfd, &readfds)) {
			buflen = real_read(clientfd, buf, BUFSIZE);
			if (buflen < 0 && errno != EPIPE) {
				logerr("read from client");
				break;
			}
			if (buflen <= 0) {
				shutdown(serverfd, 1);
				close_state |= C_TO_S;
				if (close_state == (C_TO_S | S_TO_C)) {
					break;
				}
			}
			bp = buf;
			while (buflen) {
				alarm(WRITE_TIMEOUT);
				nwritten = write(serverfd, bp, buflen);
				alarm(0);
				if (nwritten < 0 && errno != EPIPE) {
					logerr("write to server");
					break;
				}
				if (nwritten <= 0) {
					real_shutdown(clientfd, 0);
					close_state |= C_TO_S;
					if (close_state == (C_TO_S | S_TO_C)) {
						break;
					}
				}
				buflen -= nwritten;
				bp += nwritten;
			}
		}
		if (FD_ISSET(serverfd, &readfds)) {
			buflen = read(serverfd, buf, BUFSIZE);
			if (buflen < 0 && errno != EPIPE) {
				logerr("read from server");
				break;
			}
			if (buflen <= 0) {
				real_shutdown(clientfd, 1);
				close_state |= S_TO_C;
				if (close_state == (C_TO_S | S_TO_C)) {
					break;
				}
			}
			bp = buf;
			while (buflen) {
				alarm(WRITE_TIMEOUT);
				nwritten = real_write(clientfd, bp, buflen);
				alarm(0);
				if (nwritten < 0 && errno != EPIPE) {
					logerr("write to client");
					break;
				}
				if (nwritten <= 0) {
					shutdown(serverfd, 0);
					close_state |= S_TO_C;
					if (close_state == (C_TO_S | S_TO_C)) {
						break;
					}
				}
				buflen -= nwritten;
				bp += nwritten;
			}
		}
	}
}

void
alarm_handler(sig)
	int	sig;
{
	exit(0);
}

static void
logmsg(char *msg)
{
	if (isdaemon) {
		syslog(LOG_ERR, msg);
		return;
	}
	fprintf(stderr, "transocks[%d]: %s\n", getpid(), msg);
}

static void
logerr(char *msg)
{
	char	tmp[2048];

	sprintf(tmp, "%s: %s", msg, strerror(errno));
	logmsg(tmp);
}
